// 
//  C3 for C4 program.
// We no longer need most of the logical information
// since the cubie colors are determined by the parent C4 cube.
// Also, the motion, etc., of the cubies is taken care of by the parent.
// So the cubie stuff has been gutted.

//		Double buffering is implemented for smooth drawing.  Thanks to O'Reilly for 
//	showing me how.
//
//	Conventions: 
//  Corners on the cube are numbered from 0 (1,1,1) 1 (1,1,-1) 2 (1,-1,2) to 7 (-1,-1,-1).
//    (Change 1 to 0 and -1 to 1 and interpret as a binary number.)
//
//  Cubies in general are numbered from 0 (111) 1 (110) 2 (11-1) to 26 (-1-1-1).
//    (Change 1 to 0, 0 to 1 and -1 to 2 and interpret as a ternary number.)
//
//		Given a face, we have a local coordinate system indexing from 0 to 2
//	for rows and columns.  These we convert to and from absolute cube locations
//	with functions like FCtoIn (Face coordinates to cube indexing.) and FCtoCorIn
//  for corners
//
//		The orientations for these faces is determined by the usual right hand rule
//	for the three axes so that we have an "up" and "left" from the "other two axes",
//	those that are not given by which face we're examining.  Notice that the vertical
//  axis is given by the first number (j) in face coords, but second in abs coords 
//  (etc.)
//
// 
// Major change, 3/21.  The C3 cube will actually handle both faces on a
// given axis.  It will draw two cubes.  It will use a single orientation
// matrix for rendering.  It will keep track of +1 and -1 faces.
// It needs to know its axis number, but not face numbers (since it's doing
// both of them.
//

import java.applet.Applet;
import java.awt.*;
import java.awt.Graphics;
import java.awt.Color;
import java.awt.Event;
import java.awt.event.*;
import java.util.Vector;
import java.io.*;
import java.io.InputStream;
import java.io.IOException;
import java.net.URL;
import java.util.EventListener;
import java.lang.Math;

// Changes for C3forC4 cube:
//		added C4Face to record which face "this" D3 cube is.
//		Changed color determination for drawing cubie:  Use C4LCube
//				to tell us color.  Just call that function.
class BasicCubeInfo	{

	// Convert a D3 vector to an index for a 3x3x3 array.
	// The point in v has values between 0 and 2.   111 is 0;  -1-1-1 is 27
	static int VtoIn (D3iVector v) {  
		return 9*(1-v.getValue(0)) + 3*(1-v.getValue(1)) + 1-v.getValue(2);
		}
	
	static int VtoIn (int x, int y, int z) {  
		return 9*(1-x) + 3*(1-y) + 1-z;
		}
	
	// Convert a corner vector to an index between 0 and 7
	// 111 is 0, -1-1-1 is 7
	static int CVtoIn (D3iVector v) { 
		return 2*(1-v.getValue(0)) + (1-v.getValue(1)) + (1-v.getValue(2))/2;
		}
		
	// D3iVector to face number.  This requires the vector to 
	// have two 0 values and one value of +/- 1.
	//
	static int VtoFN(D3iVector v) { 
		int val;
		for (int i=0; i<3; i++)  {
			val = v.getValue(i);
			if (val == 1) return i;
			else if (val == -1) return i+3;
			}
		return -1; // will cause exception if (some) bad vector is passed.
		}

	// return the face number for the given axis and +/- 1 value.
	// no error checking.
	static int VtoFN(int index, int value) {
		value = (1-value)/2;
		return index + value*3;
		}

	// Face coordinates to vector:  In is [0to5][0-2][0-2]  Out is -1 to 1, 3 times
	static void FCtoV(int face,int j, int k, D3iVector target)  {
		int axis = face%3;
		int side = face<3? 1 : -1;

		j = 1-j; k = k-1;  // Adjust to cubie coords
		int x = side,y=k,z=j;  // use x0 for axis as default
		
		if (axis == 1) {
			x=j; z=k; y=side;
			}
		else if (axis==2) {x=k; y=j; z=side;}
				
		target.setValue(x,y,z);		
		}

	// Given a face, and face coordinates j and k, calculate the index of the cubie
	// that lives in that position.  Combines FCtoV and VtoIn.  Use right hand rule
	// to determine which axis is where.  Note j is row (vertical coord) 
	// and k is column  (horiz coord).
	// Note, however, that we do not take into account which side we're on when
	// calculating orientation.
	static int FCtoIn(int face, int j, int k)  {
		int axis = face%3;
		int side = face<3? 1 : -1 ;
		j = 1-j; k = k-1;  // Adjust to cubie coords.  Note 0 vert is top;0 horiz is left.
		int x = side,y=k,z=j;  // use x0 for axis as default
		
		if (axis == 1) {
			x=j; z=k; y=side;
			}
		else if (axis==2) {x=k; y=j; z=side;}
				
		return 9*(1-x) + 3*(1-y) + 1-z;
		}  /// end of FC to Index converstoin


	// FCtoCorIn.  As above, but returns a corner index instead of a cubie index.
	static int FCtoCorIn(int face, int j, int k)  {
		int axis = face%3;
		int side = face<3? 1 : -1 ;
		j = 1-j; k = k-1;  // Adjust to cubie coords.  Note 0 vert is top;0 horiz is left.
		int x = side,y=k,z=j;  // use x0 for axis as default
		
		if (axis == 1) {
			x=j; z=k; y=side;
			}
		else if (axis == 2) {x=k; y=j; z=side;}
		
		return 2*(1-x) + (1-y) + (1-z)/2;
		}  /// end of FC to Index converstoin

}  // basic cube info


// The visual aspects of the cube.
// Use visual 3 space here.
//
// New visual aspect of cube approach.
// Each axis is assigned a frame.
// The frame is implemented in this VCube class.
// These frames handle the visual orientations of the
// faces.    
class VCube extends BasicCubeInfo  {

	static boolean DEBUG = true;
	static int margin = 3; // for offsets, # of pixels from edge of rectangle



////////////////////////////////////////////////////
////////////////////////////////////////////////////
//
//
// Start of C3 visual handling for C4 cube.
//
//
////////////////////////////////////////////////////			

	protected int C4Face; // The number of this face, 0..7.
	protected int C4Axis; // The axis for this face,  0..3.
	protected boolean ViewPlus;  // Do we view from plus or minus 1?
	protected int HideAxis;  // Hide this axis from view
	
	protected int half; // For finding center of window
	protected float max;  // for calculating coords in window.
	protected int cx, cy;  // center coords of window
	protected D3Matrix orient;  // The visual orientation of this cube.
			// the x, y and z axes of cube are stored as column vectors, 
	
	protected int [] face; // the last faces drawn
	protected int colAxis; //  The axis to collapse on when drawing the cube.
	
	protected C4VCube parent;  // the parent of this C3 cube face.
	
	protected Point [] corpoints; //  coords of all corner points on cube for display.
	protected Point [][] FacePoints;  // 4 x 4 array of points on each face
	protected Point [] ThisFaceCorners;  // For generating the above array, keep track of
					// this face's corners.	
					
	protected Polygon [][][] ThePolys; // Holds the 27 polygons used for the
			// currently visible faces.  Used to determine which polygon
			// a point lies in.
			
			
	VCube(int myFace, C4VCube C4C) {  // Default constructor
		C4Face = myFace;
		C4Axis = myFace%4;
		colAxis = 2;
		parent = C4C;
		
		HideAxis = C4Axis;
		ViewPlus = myFace < 4;
		setLocation(0,0,200);

		// A face has 9 polys.
		corpoints = new Point[8]; //  coords of points for display.
		FacePoints = new Point [4][4];  // Points on the face at corners of cubies
		
		ThePolys = new Polygon[3][3][3]; // Face by row by column
		
		face = new int [3];
		
		orient = new D3Matrix();
		
		if (C4Face>3) 
			orient.xrot(180);  // just for fun.

		int i,j,k;
		
		// Get the points for various arrays.
		for (i=0; i<8; i++)
			corpoints[i] = new Point();
			
		for (j=0; j<4; j++)
			for (k=0; k<4; k++)
	 			FacePoints[j][k] = new Point();

		for (i=0; i<3; i++)
			for (j=0; j<3; j++)
				for (k=0; k<3; k++)
		 			ThePolys[i][j][k] = new Polygon();

	} // end constructor 
	
	// Constructor with sizes, etc.
	VCube(int myFace, C4VCube C4C, int x, int y, int size) {
		this( myFace, C4C);
		setLocation(x,y,size);
		}

	public void clearPoly (Polygon P) {
		P.npoints = 0;
		}

	public void setVisInfo (int a) {
		HideAxis = a;
		}
		
	public void setViewPlus(boolean p) {
		ViewPlus = p;
		}
		
	public void setLocation(int x, int y, int size) {
		half = size/2; // center in window.
		cx = x + half; cy = y + half;  // center coordinates
		max =  (half  - margin)/(float)1.6;  // for calculating coords in window.
					// that's root 3 on the bottom.
		}

	public void setOrientation(D3Matrix M) {
		orient.setValue(M);
		}

	public void VRotate(int n, boolean horiz) {
		if (horiz) orient.yrot(5);
		else orient.xrot(5);
	
	}
	
	// This function takes coordinates of a point as arguements
	// will return a cubie index if the point is on the cube.
	// If not, it returns -1.
	// It's pretty simple:  We just check the 27 polygons from the 
	// last rendering and if contains is true for any of them, we
	// call the parents FCtoIn function.
	// 
	public int getCubieIndex(int x, int y) {
		int i,j,k,idx;
		D3iVector temp = new D3iVector();
		for (i=0; i<3; i++)
			for (j=0; j<3; j++)
				for (k=0; k<3; k++)
					if (ThePolys[i][j][k].contains(x,y)) {
						// get coords of this cubie in this cube into temp
						FCtoV(face[i],j,k,temp); 
						idx = parent.FCtoIn(C4Face, temp.x(), temp.y(), temp.z());
						if (parent.DEBUG) System.out.println("Cubie index found:  " + idx);
						return idx;
						
					}
		return -1;
	}  // end getCubieIndex
	
	
	// Render the  cube.  Given graphics context, upper left corner
	// and width/height of (square) window
	// Mommy is the parent C4 cube that we use to determine
	// the colors of the cubies.
	synchronized public void RenderCube(Graphics g) {

		int i;
		
		// First, find out which faces are visible:
		for (i=0;i<3; i++) 
			if (orient.getValue(2,i) > 0)  // draw positive face for this axis
				face[i] = i; 
			else ///  Draw negative face
				face[i] = i+3; 

		// Get visual corners for all 8 corners
		CalculateCorners();
		
		// find the 27 polygons for this cube.
		CalculatePolygons();
		
		// render the polygons
		RenderFaces(g);
		
		// draw the orientation widget for this face
		DrawOrient(g, 20, 25, cy+half);
		
		// indicate the face
		g.drawString("Face:  x" + C4Axis + " = " + String.valueOf(C4Face/4==0?1:-1),
				cx-20, cy + half + 35);
	}  // end render cube

	synchronized private void CalculatePolygons() {
	
		int i,j,k;   //, index;
				
		// i is face; j is row; k is column.
		for ( i = 0; i<3; i++)  {
			
			// Get the 4x4 points on a given face.
			CalculateFacePoints(i);

			// Now generate the 9 polygons for this face
			for (j=0; j<3; j++) 
				for (k=0; k<3; k++) {
					// Set up the polygon
	//					p = new Polygon();  
					clearPoly(ThePolys[i][j][k]);
					addPoint(ThePolys[i][j][k],FacePoints[j][k]);
					addPoint(ThePolys[i][j][k],FacePoints[j][k+1]);
					addPoint(ThePolys[i][j][k],FacePoints[j+1][k+1]);
					addPoint(ThePolys[i][j][k],FacePoints[j+1][k]);
				}
			}
			
	}
	
	synchronized private void RenderFaces(Graphics g) {
		int i,j,k;
		D3iVector temp = new D3iVector();
		int cubieIndex;
		
		for (i=0;i<3; i++) {
					
			// Now draw the 9 polygons for this face
			for (j=0; j<3; j++) 
				for (k=0; k<3; k++) {

					// get coords of this cubie in this cube into temp
					FCtoV(face[i],j,k,temp); 
					
					cubieIndex = parent.FCtoIn(C4Face, temp.x(), temp.y(), temp.z());
					
					// Draw the polygon using parent C4Cube to choose color		
					g.setColor(parent.CubieToColor(C4Face, cubieIndex, face[i]));
					g.fillPolygon(ThePolys[i][j][k]);
					if (!parent.isCubieSelected(cubieIndex)) {
						g.setColor(Color.black);
						g.drawPolygon(ThePolys[i][j][k]);
						}
					else {  // highlight selected polygons.  Not ideal....
						g.setColor(Color.red);
						g.drawPolygon(ThePolys[i][j][k]);
						ThePolys[i][j][k].translate(1,1);
						g.setColor(Color.black);
						g.drawPolygon(ThePolys[i][j][k]);
						ThePolys[i][j][k].translate(-2,-2);
						g.drawPolygon(ThePolys[i][j][k]);
						ThePolys[i][j][k].translate(1,1);
						}
						
						
					} 
			}  // end drawing faces.
	
	
	}
	
	// This function will draw a little 3d axis representation
	// cx and cy are "origin" of pic
	public void DrawOrient(Graphics g, int length, int cx, int cy)  {
		// We just take the first two values in the columns
		// of the orientation matrix (projections of multiplying orient on
		// unit vectors)
		int i, px, py,j;
		for (i=0; i<3; i++) {
			px = cx + (int) (length*orient.getValue(0,i));
			py = cy - (int) (length*orient.getValue(1,i));
			g.setColor(Color.black);
			g.drawLine(cx,cy,px,py);
			j = i<C4Axis? i: i+1;
			g.drawString("x" + j,px,py);
			
			}
		
		} // end draw orient
		
		
	// Calculate the screen coords for the points on a face.
	// PRECONDITION:  Face Point's corners are already set.  
	// Interpolate the rest of the points  
	private void CalculateFacePoints(int i ) {
		
		int j,k;

		// First, copy the corner points into the face point matrix.
		// Put the four corners in the face points
		for (j = 0; j<2; j++)
			for (k=0; k<2; k++){
				// Note 4x4 points, but 3x3 cubies.
				copyPoint(FacePoints[3*j][3*k], 
						corpoints[FCtoCorIn(face[i],2*j,2*k)]); // dest = source
				}
					
		
		
		// Now calculate the rest of the face points via interpolation.
		for (j=0; j<4; j++) { // Do the jth row of points.
			// First to left, right sides if not on corner
			// two points to interpolate into destination with scale factor
			if (j >0 && j<4) {
				interpolate(FacePoints[0][0],FacePoints[3][0],FacePoints[j][0] ,0.333333*j);
				interpolate(FacePoints[0][3],FacePoints[3][3],FacePoints[j][3] ,0.333333*j);
				}
			// Do the middle points
			interpolate(FacePoints[j][0],FacePoints[j][3],FacePoints[j][1],0.333333);
			interpolate(FacePoints[j][0],FacePoints[j][3],FacePoints[j][2],0.666666);
			}			
		} // end calculate face points
		
	
	// Calculate the visual coordinates of the 8 corners
	private void CalculateCorners() {
		D3Vector temp = new D3Vector();
		int i,j,k, index=0;
		int xval, yval;
		for ( i=1; i>-2; i-=2)  
			for ( j=1; j>-2; j-=2) 
				for ( k=1; k>-2; k-=2) {
					temp.setValue(i,j,k);
					orient.multSetV(temp);
					xval = (int) (temp.getValue(0) *  max);
					yval = (int) (temp.getValue(1) *  max);
					corpoints[index++].setLocation(xval+cx,cy-yval);
					}
	}  // end calculate corners
	
		 
	// Point manipulation functions.
	//
	// Interpolate points with weighting factor.  
	// factor should be between 0 and 1
	
	private void interpolate(Point p1, Point p2, Point Target, double factor) {
		Target.x = p1.x + (int) ((p2.x - p1.x)*factor);
		Target.y = p1.y + (int) ((p2.y - p1.y)*factor);
		}
	public void addPoint(Polygon p, Point pt) {  // Why isn't this a polygon function?
		p.addPoint(pt.x, pt.y);
		}
	public void copyPoint(Point dest, Point source) {  // should just use clone()
		dest.setLocation(source.x, source.y);
		}



}  // end of class visual cube.



// "Rendered" obsolete by RenderFaces.
/*
	// Render a face of the cube
	// PRECONDITION:  the screen coords are stored in the 4x4 FacePoints.
	private void RenderFace(Graphics g, int face)   {

		Polygon p;
		int i,j,k;
		D3iVector temp = new D3iVector();
		
		// Now generate and draw the 9 polygons for this face
		for (j=0; j<3; j++) 
			for (k=0; k<3; k++) {
				// Set up the polygon
				p = new Polygon();  
				addPoint(p,FacePoints[j][k]);
				addPoint(p,FacePoints[j][k+1]);
				addPoint(p,FacePoints[j+1][k+1]);
				addPoint(p,FacePoints[j+1][k]);
				
				// get coords of this cubie in this cube into temp
				FCtoV(face,j,k,temp); 
				
				// Draw the polygon using parent C4Cube to choose color				
				g.setColor(parent.CubieToColor(C4Face,temp.x(), temp.y(), temp.z(),face));
				g.fillPolygon(p);
				g.setColor(Color.black);
				g.drawPolygon(p);
				} 
		
		}  // end RenderFace
*/
	

