//
// C4 Cube classes.
// Based on C3 development.
//
// Hypercube, rubik's cube.  
// Dan Talayco
// March, 1998
// 
// 
// Basic idea:  C4 has 8 faces each of which is a 3D cube.  Notice,
// though, that a layer is a face drawn through another dimension,
// so it's actually a 4D object.
//
// Structure of program:
//		Basic cube information and indexing functions are set up in
// 	an abstract class.  This is extended by "cubie", the small subunits
//	of the cube, as well as LCube, the logical information about the cube.
//	"cubie" does most of the work for rotating layers of the cube, keeping
//	track of its orientation and how far it's strayed from home.
//		LCube coordinates the information about the cube keeping an array of
//  the cubies as well as a second array that tracks which cubie is in which
// 	location.  
//
//		We differentiate between:
// 			*** Absolute 3 space which is static, set at the center of the cube
// 			and gives us the coordinates of a location for a cubie to sit.
//			*** Visual 3 space which is what the cube can move in to show different
//			sides to the viewer.  We project from visual 3 space to the xy axis 
//			looking from above to determine how the cube appears to the viewer.
//			*** Local cubie 3 space which is the orientation of an individual cubie.
//			This basically answers the question:  "Which cubie face is pointing in
//			a given absolute direction?
//		Displaying is what is computationally intensive, so
//	we actually do (simple, integer) matrix multiplication to update orientations
//	and positions of cubies.  This seems to work well.  
//		Calculating face points could be optimized to run faster.  That's the first
// 	attack for getting animation to work well.
//
//		Double buffering is implemented for smooth drawing.  Thanks to O'Reilly for 
//	showing me how.
//
//	Conventions: 
//  Corners on the cube are numbered from 0 (1,1,1,1) 1 (1,1,1,-1)... to 15 (-1,-1,-1,-1).
//    (Change 1 to 0 and -1 to 1 and interpret as a binary number.)
//
//  Cubies in general are numbered from 0 (1111) 1 (1110) 2 (111-1) to 81 (-1-1-1-1).
//    (Change 1 to 0, 0 to 1 and -1 to 2 and interpret as a ternary number.)
//
//		Given a face, we have a local coordinate system (a D3cube) indexing from -1 
//  to 1 in 3 dims.  These we convert to and from absolute cube locations
//	with functions like FCtoIn (Face coordinates to cube indexing.) and FCtoCorIn
//  for corners
//
//		The orientations for these faces is determined by the usual right hand rule
//	for the three axes so that we have an "up" and "left" from the "other two axes",
//	those that are not given by which face we're examining.  Notice that the vertical
//  axis is given by the first number (j) in face coords, but second in abs coords 
//  (etc.)
//
// In LCube: 
// TheCubies:  array of 81 cubies indexed by their original location. 
// WhichWhere:  Array of 81 pointers to the cubies above, indexed by location, but
// 		pointing to the cubie that occupies that position currently.  
// 

import java.applet.Applet;
import java.awt.*;
import java.awt.Graphics;
import java.awt.Color;
import java.awt.Event;
import java.awt.event.*;
import java.util.Vector;
import java.io.*;
import java.io.InputStream;
import java.io.IOException;
import java.net.URL;

class BasicC4CubeInfo {
	static boolean DEBUG = true;
	static Color purple = new Color(150, 0, 200);
	static Color orange = new Color(255, 150, 0);
	static Color darkgray = new Color(50,50,50);
	static Color AVeryNiceBlue = new Color(0, 200, 200);
	
	static Color [] facecolor = {Color.red, Color.yellow, Color.blue, Color.green, 
		purple, orange, Color.white, darkgray};  // add white and gray for 4d cube
	static int margin = 3; // for offsets, # of pixels from edge of rectangle
	
	static D4iMatrix [][] FaceRotations;  // The 90 degree rotations about each face.
	
	static {
		int i,j;
		FaceRotations = new D4iMatrix[4][4];
		for ( i=0; i<4; i++)
			for (j=0; j<4; j++)  // We never use the i,i rotations
				FaceRotations[i][j] = new D4iMatrix(i,j); // init to face rotation.
							// NOTE:  i and j are the AXES THAT MOVE.
		if (DEBUG) // Display the rotations
			for ( i=0; i<4; i++)
				for (j=0; j<4; j++)  // We never use the i,i rotations
					System.out.println(FaceRotations[i][j]);
		}

	// Convert a D4 vector to an index for a 4x4x4 array.
	// The point in v has values between 0 and 2.   111 is 0;  -1-1-1 is 80
	public int VtoIn (D4iVector v) {  
		return 27*(1-v.getValue(0)) + 9*(1-v.getValue(1)) + 3*(1-v.getValue(2))
				+ 1-v.getValue(3);
		}
	
	public int VtoIn (int x0, int x1, int x2, int x3) {  
		return 27*(1-x0) + 9*(1-x1) + 3*(1-x2) + 1-x3;
		}
	
	// Convert a corner vector to an index between 0 and 7
	// 111 is 0, -1-1-1 is 7
	public int CVtoIn (D4iVector v) { 
		return 4*(1-v.getValue(0)) + 2*(1-v.getValue(1)) + (1-v.getValue(2)) +
				(1-v.getValue(3))/2;
		}
		
	// D4iVector to face number.  This requires the vector to 
	// have two 0 values and one value of +/- 1.
	//
	public int VtoFN(D4iVector v) { 
		int val;
		for (int i=0; i<4; i++)  {
			val = v.getValue(i);
			if (val == 1) return i;
			else if (val == -1) return i+4;
			}
		return -1; // will cause exception if (some) bad vector is passed.
		}

	// return the face number for the given axis and +/- 1 value.
	// no error checking.
	public int VtoFN(int index, int value) {
		value = (1-value)/2;
		return index + value*4;
		}

	// Face coordinates to vector:  In is [0to7][1 to -1 3 times]  Out is -1 to 1, 4 times
	public void FCtoV(int face, int x, int y, int z, D4iVector target)  {
		int axis = face%4;
		int side = face<4? 1 : -1;
		
		int x0 = side, x1 = x, x2 = y, x3 = z;
		
		if (axis == 1) {x0 = x; x1 = side;}
		else if (axis == 2) {x0=x; x1=y; x2=side;}
		else if (axis == 3) {x0=x; x1=y; x2=z; x3=side;}
				
		target.setValue(x0,x1,x2,x3);		
		}

	// Given a face, and face coordinates x,y,z, calculate the index of the cubie
	// that lives in that position.  Combines FCtoV and VtoIn.
	// Note, however, that we do not take into account which side we're on when
	// calculating orientation.
	public int FCtoIn(int face, int x, int y, int z)  {
		int axis = face%4;
		int side = face<4? 1 : -1 ;

		int x0 = side, x1 = x, x2 = y, x3 = z;
		
		if (axis == 1) {x0 = x; x1 = side;}
		else if (axis == 2) {x0=x; x1=y; x2=side;}
		else if (axis == 3) {x0=x; x1=y; x2=z; x3=side;}
				
		return 27*(1-x0) + 9*(1-x1) + 3*(1-x2) + 1-x3;
		}  /// end of FC to Index conversion


	// FCtoCorIn.  As above, but returns a corner index instead of a cubie index.
	public int FCtoCorIn(int face, int x, int y, int z)  {
		int axis = face%4;
		int side = face<4? 1 : -1 ;

		int x0 = side, x1 = x, x2 = y, x3 = z;
		
		if (axis == 1) {x0 = x; x1 = side;}
		else if (axis == 2) {x0=x; x1=y; x2=side;}
		else if (axis == 3) {x0=x; x1=y; x2=z; x3=side;}

		return 4*(1-x0) + 2*(1-x1) + (1-x2) + (1-x3)/2;
		}  /// end of FC to corner Index conversion
		
	}  // End of class, basic cube info.


class C4cubie extends BasicC4CubeInfo {
	protected D4iMatrix or;  // the orientation:  Rotation of identity through 90 degree rots
	protected  D4iVector loc; // location of cubie. position indicated in absolute 4 space
			// Entries are 1,0 or -1
			
	protected D4iVector initloc;  // Initial location of this cubie.
	protected D4iVector faces;  // Which face is pointing in x0, x1, x2, x3 direction.
			// holds numbers from 0 to 7.
	protected boolean iscorner;  // true if all coords are +/-1
	protected int myIndex;  // The index of this cubie.
	
	C4cubie(D4iVector myloc) { 
		or = new D4iMatrix();// init to unit matrix
		initloc = new D4iVector(myloc); // init location to that passed.
		loc = new D4iVector(myloc); // init location to that passed.
		faces = new D4iVector(0,1,2,3);
		
		myIndex = VtoIn(myloc);
		
		iscorner = true;
		for (int i=0;i<4; i++)
			if (loc.getValue(i) == 0) iscorner = false;
		}
	
	// Reset the cubie.
	public void ResetCubie(D4iVector myloc) {
		or.unit();
		initloc.setValue(myloc);
		loc.setValue(myloc);
		faces.setValue(0,1,2,3);

		iscorner = true;
		for (int i=0;i<4; i++)
			if (loc.getValue(i) == 0) iscorner = false;
		}
				
	// Given the face of a cube, look at its orientation and 
	// decide which color belongs on that face.
	public Color WhichColor (int face) {
		if (face < 4) return facecolor[faces.getValue(face)];
		else return facecolor[(faces.getValue(face-4) + 4) % 8];
		}
		
	// This function will move the cubie given the 
	// face and direction of the parameters.
	// This is specific to a 4 D layer movement.
	// AXIS1 AND AXIS2 ARE THE AXES THAT MOVE.
	public void movecubie (int axis1, int axis2, boolean ccw) {
		D4iVector arow = new D4iVector();

		if (ccw) {
			FaceRotations[axis1][axis2].multSetV(loc);
			or.Lmult(FaceRotations[axis1][axis2]);
		}
		else {
			FaceRotations[axis2][axis1].multSetV(loc);
			or.Lmult(FaceRotations[axis2][axis1]);
		}

		// Now alter the face color vector.  Rows of orientation matrix determine face.
		for (int i=0; i<4; i++) {  
			or.getRow(i,arow);
			faces.setValue(i, VtoFN(arow));
			}
		} // end of move cubie.
	
	public boolean ishome() {  // NOTE:  It is possible that 
	// the orientation may not be the identity, but the proper face
	// is showing for the center cubies.  This is a strange thing about
	// RC.
		return (loc.equals(initloc) && or.isunit());
		}
	public String toString() {
		return " H:" + initloc + "C:" + loc+"F:"+faces+"\n"+or;
		}
	public String locString() {
		return " From: " + initloc + "Now:" + loc+".";	
	}
} // end of class cubie


	
class C4LCube extends BasicC4CubeInfo { // Logical cube info

	// Faces numbered from 0 to 7 are x0, x1,x2,x3,-x0,...
	// corners are numbered 0 to 15; use grey code on +/- rep with 0 being 1111
	// 1 being 111 -1, etc.  Use 1 as 0 and -1 as 1 and interpret as binary.


//	D4iVector [] corners;
	protected C4cubie [] TheCubies; // The cubies in the cube.
	protected C4cubie [] WhichWhere;  // Pointers to above array keeping track of which cubie is
		// in which location.
		// This is not optimal for time considerations....
		
	C4LCube () {
		int i, j, k, m, index=0;
		int ind;
		D4iVector where=new D4iVector(); // for placing cubies

		TheCubies = new C4cubie[81];
		WhichWhere = new C4cubie[81];
//		corners = new D4iVector[8];
		
		// Initialize the cubies
		for (i=1; i>-2; i--)
			for (j=1; j>-2; j--)
				for (k=1; k>-2; k--) 
					for (m=1; m>-2; m--) {
						where.setValue(i,j,k,m);
						ind = VtoIn(where);
						TheCubies[ind] = new C4cubie(where);
						WhichWhere[ind] = TheCubies[ind];
						}  // setting cubies up.

	}  // end C4LCube constructor
	
	public void ResetCube() {
		int i,j,k,m,ind;
		D4iVector where=new D4iVector(); // for placing cubies
		// Initialize the cubies
		for (i=1; i>-2; i--)
			for (j=1; j>-2; j--)
				for (k=1; k>-2; k--) 
					for (m=1; m>-2; m--) {
						where.setValue(i,j,k,m);
						ind = VtoIn(where);
						TheCubies[ind].ResetCubie(where);
						WhichWhere[ind] = TheCubies[ind];
						}  // setting cubies up.
		}  // end of reset cube

	// This function will set the WhichWhere pointers to point to
	// the appropriate cubie; namely, the cubie in the array TheCubies
	// that currently lies in that position.
	public void adjustWWCubies ()  {
		int i;
		int myPosition;
		boolean check[];
		
		check = new boolean [81];
		
		for (i=0;i<81;i++) {
			myPosition = VtoIn(TheCubies[i].loc);
			WhichWhere[myPosition] = TheCubies[i];
			
			// tracking integrity of cubie map.
			if (DEBUG) 
				if (check[myPosition]) 
					System.out.println("Error adjusting which where: " + i + " "
						+ myPosition);
				else check[myPosition] = true;
				
			}
		}  // end of adjust ww cubie matrix.
		
	// This important function will take a description of a cubie by
	// face and D3Cube coordinates, and a face in which we're looking
	// and return the color of the face of the referenced cubie facing
	// in the direction facedir.
	public Color CubieToColor (int d4face, int i, int j, int k, int d3face) {
		// First, convert to a cubie index.
		int index = FCtoIn(d4face, i,j,k);
		// Calculate the D4 face direction from face and d3face:
		int facedir = d3tod4face(d4face,d3face);
		// Then call that cubies color function.
		return WhichWhere[index].WhichColor(facedir);
		}

	public Color CubieToColor (int d4face, int cubieIndex, int d3face) {
		// Calculate the D4 face direction from face and d3face:
		int facedir = d3tod4face(d4face,d3face);
		// Then call that cubies color function.
		return WhichWhere[cubieIndex].WhichColor(facedir);
		}

	// Given a d4 face and a d3 face, find the d4 direction for the d3 face.
	// It might be better to reverse the orientation for faces when the
	// d4 face is on a negative side.  But it's a little unclear.  This should work fine.
	public int d3tod4face(int face, int d3face) {
		int low = face%4, high = low+4;
		if (d3face < low) return d3face;
		else if (d3face < high) return d3face+1;
		else return d3face+2;
		}


	
	// Rotate a face of the cube
	
	public synchronized void RotateFace(int face, int axis1, int axis2, boolean CCW)  {  
		// For each of the cubies on the face
		// apply the move cubie operation.  
		int i, j, k;

		for (i=1; i>-2; i--)
			for (j=1; j>-2; j--)
				for (k=1; k>-2; k--) 
					WhichWhere[FCtoIn(face,i,j,k)].movecubie(axis1, axis2, CCW);

		adjustWWCubies(); // Change the cubie map
					
		}  // end of rotate face 

	public String toString() {
		String temp = new String();
		int i,j;
		for ( i=0;i<9;i++)  {
			for (j=0;j<9;j++) 
				temp += WhichWhere[i*3+j];
			temp += "\n";
			}
		return temp;
		}
} // end logical aspects of cube


// The visual aspects of the cube.
// Use visual 3 space here.
// Visual Orientation is stored in the faces.
// But the cubie positions, etc., need to be translated down.
//
class C4VCube extends C4LCube 
	//implements MouseMotionListener, ActionListener  
	{

	protected FaceViewer TheVFaces [];
	protected int FVhoriz, FVvert; // horizontal and vertical extents of an axis viewer
	protected boolean cubieSelected[];  // if a cubie is clicked, it becomes selected.
	
	C4VCube( ) {
		super();
		MyInit(0,0,150,200);
	}			
	// First viewer is placed at top left, with indicated
	// horizontal and vertical extents.
	// Face viewers are then tiled:
	//   0 1 2 3 
	//   4 5 6 7
	//
	C4VCube(int l, int t, int h, int v) {
		super();
		if (DEBUG) System.out.println("Initializing  C4VCube.");
		MyInit (l,t,h,v);
		System.out.println("C4VCube initialized.");
	

	} // end constructor 
	
	protected void MyInit(int l, int t, int h, int v) {
		int i;
		
		FVhoriz = h; FVvert = v;
		cubieSelected = new boolean [81];
		TheVFaces = new FaceViewer[8];
		for (i=0;i<8;i++)  {
			TheVFaces[i] = new FaceViewer(i, FVhoriz, FVvert, this);
//			TheVFaces[i].show();
			}		
		
	}  // end init function

	public Panel FacePanel(int i) {
		return TheVFaces[i];
		}

	public void paint(Graphics gc)  {
		int i;
		for (i=0; i<8; i++)
			TheVFaces[i].RenderFace(gc);
		gc.drawLine(0,FVvert, 4*FVhoriz,FVvert);
	}
	
	public void ShowInfo()  {
    	System.out.println("This is a prototype.  Release C4 Ver. 0.3a for Java 1.1");
    	System.out.println("You are looking at the eight faces of a ");
    	System.out.println("four dimensional Rubik's cube.");
    	System.out.println("Click on a cube to highlight a cubie (small cube).");
    	System.out.println("That cubie will be highlighted in other faces.");
    	System.out.println("Click the cubie again to deselect it.  You can highlight");
    	System.out.println("more than one cubie at a time.");
    	System.out.println("Click close to a cube to change the visual orientation");
    	System.out.println("of that face.  This does not affect the cube.");
    	System.out.println("Click a button below a face to");
    	System.out.println("rotate that face (layer) about the selected axis.");
    	System.out.println("Please report successes and failures to");
    	System.out.println("talayco@iName.com");
	}

	// This function changes the state of a cubie selection;
	// when a cubie is clicked, it is selected.
	public void cubieClicked(int cubieIndex) {
		cubieSelected[cubieIndex] = !cubieSelected[cubieIndex];
		RenderCube();
	}
	
	public boolean isCubieSelected(int cubieIndex) {
		return cubieSelected[cubieIndex];
	}
	
	// Render the  cube.  Given graphics context, upper left corner
	// and width/height of (square) window
	public void RenderCube() {

		int i;
		for (i=0; i<8; i++)
			TheVFaces[i].repaint();

	}  // end render cube
	
	// reset the cube visually
	public void VReset() {
		int i;
		for (i=0; i<8; i++)
			TheVFaces[i].VReset();
		RenderCube();
	}
	// Perform a visual rotation for the given face (0 to 7) and degrees
	public void VRotate(int face, int xdegrees, int ydegrees) {
		TheVFaces[face].VisRotate( true, xdegrees);
		TheVFaces[face].VisRotate( false, ydegrees);
	}
	
	// rotat all faces with this signature
	public void VRotate(boolean horiz, int degrees) {
		int i;
		if (horiz)
			for (i=0;i<8; i++)
				TheVFaces[i].VisRotate(true, degrees);
		else
			for (i=0;i<8; i++)
				TheVFaces[i].VisRotate( false, degrees);
	}

	void quit() {
		int i;
		for (i=0; i<8; i++)
			TheVFaces[i].stop();

	}  // end quit


}  // end of class visual cube.


